# listener.py
# JAB 4/9/07

import numpy
import os
import socket
import struct
import sys
import threading
import time
import wx

class CallbackThread(threading.Thread):
    def __init__( self, panel_com, port, callback, datafile, parse_packet ):
        """Initialize callback thread."""
        threading.Thread.__init__( self )
        self.init_successful = False
        self.panel_com = panel_com
        self.port = port
        self.callback = callback
        self.datafile = datafile
        self.parse_packet = parse_packet

        self.sockobj = socket.socket( socket.AF_INET, socket.SOCK_DGRAM )
        try:
            self.sockobj.bind( ('localhost', self.port) )
        except socket.error:
            wx.MessageBox( "Can't bind listener port", "Error", wx.ICON_ERROR )
            return
        self.sockobj.settimeout( None )

        self.packets_received = 0
        self.init_successful = True

    def run( self ): # thread dies when this function returns
        """Thread's main function."""
        while True:
            packet = self.sockobj.recv( 4096 ) # binding call
            if packet == "terminator packet": break
            self.packets_received += 1
            parsed_data = self.parse_packet( packet ) # interface-specific
            worked_data = self.callback( parsed_data ) # user-defined function
            self.update( worked_data )
        print "saving complete"

    def update( self, data ):
        self.panel_com.SetPositions( data[0], data[1] )
        # save all data to file
        for i in range( len(data) ):
            self.datafile.write( str(data[i]) + "\t" )
        self.datafile.write( "\n" )

class Listener:
    def __init__( self, panel_com, port, parse_packet ):
        """Initialize generic listener object."""
        self.init_successful = False
        self.datafile = None
        self.port = port
        self.parse_packet = parse_packet
        
        # select callback function...
        dlg = wx.FileDialog( None, "Choose callback file", os.environ['HOME'], "", "*.py", wx.OPEN ) # Linux only
        if dlg.ShowModal() == wx.ID_OK:
            filedir = dlg.GetDirectory()
            filename = dlg.GetFilename()
            dlg.Destroy()
        else:
            dlg.Destroy()
            return
        py_dex = filename.find( '.py' )
        sys.path.insert( 0, filedir )
        try:
            callback_module = __import__( filename[:py_dex], [], [], [] )
            sys.path.pop( 0 )
        except ImportError:
            wx.MessageBox( "Can't import file\n%s"%filename, "Error", wx.ICON_ERROR )
            sys.path.pop( 0 )
            return
        # should be
        #finally:
        #    sys.path.pop( 0 )
        # ...but I get a SyntaxError (maybe python2.4?)
        callback = callback_module.callback

        # select and open data file
        dlg = wx.DirDialog( None, "Choose save directory", os.environ['HOME'], ) # Linux only
        if dlg.ShowModal() == wx.ID_OK:
            filedir = dlg.GetPath()
            dlg.Destroy()
        else:
            dlg.Destroy()
            return
        filename = time.strftime('%Y-%m-%d_%H-%M-%S' '.txt')
        fullname = os.path.join( filedir, filename )
        try: self.datafile = open( fullname, "w" )
        except IOError:
            wx.MessageBox( "Can't create file\n%s"%fullname, "Error", wx.ICON_ERROR )
            self.datafile = None
            return

        # begin callback thread: listen for and work on packets
        self.callback_thread = CallbackThread( panel_com,
                                               self.port,
                                               callback,
                                               self.datafile,
                                               self.parse_packet )
        if not self.callback_thread.init_successful:
            return
        self.callback_thread.start()
        print "saving data to", fullname

        self.init_successful = True

    def __del__( self ):
        """Cleanup open stuff, kill thread."""
        if self.datafile is not None:
            self.datafile.close()
        if self.init_successful:
            # send one last packet so that the thread will quit
            sockobj = socket.socket( socket.AF_INET, socket.SOCK_DGRAM )
            sockobj.sendto( "terminator packet", ('localhost', self.port) )

class Flytrax_listener(Listener):
    def __init__( self, panel_com ):
        Listener.__init__( self, panel_com, 28931, parse_Flytrax_packet )

class Spheretrax_listener(Listener):
    def __init__( self, panel_com ):
        Listener.__init__( self, panel_com, 28932, parse_Spheretrax_packet )
        

class DataStruct: pass # just an empty structure

def parse_Flytrax_packet( packet ):
    version = packet[0]
    data = DataStruct()
    if version == 'd':
        packet_data = struct.unpack( 'cLdff', packet )
        data.version = packet_data[0]
        data.framenumber = packet_data[1]
        data.timestamp = packet_data[2]
        data.x = packet_data[3]
        data.y = packet_data[4]
    elif version == 'e':
        packet_data = struct.unpack( 'cBLdfffffBBII', packet[:48] )
        data.version = packet_data[0]
        data.cam_no = packet_data[1]
        data.framenumber = packet_data[2]
        data.timestamp = packet_data[3]
        data.x = packet_data[4]
        data.y = packet_data[5]
        data.area = packet_data[6]
        data.slope = packet_data[7]
        data.eccentricity = packet_data[8]
        data.roi_w = packet_data[9]
        data.roi_h = packet_data[10]
        data.roi_x = packet_data[11]
        data.roi_y = packet_data[12]
        data.img = numpy.fromstring( packet[48:] )
    else:
        print "unknown packet version", version

    return data

def parse_Spheretrax_packet( packet ): 
    version = packet[0]
    data = DataStruct()
    if version == 'a':
        packet_data = struct.unpack( 'cLdfff', packet)
        data.version = packet_data[0]
        data.framenumber = packet_data[1]
        data.timestamp = packet_data[2]
        data.head_rate = packet_data[3]
        data.forw_rate = packet_data[4]
        data.side_rate = packet_data[5]
    else:
        print "unknown packet version", version
        
    return data
